#!/bin/perl

#----------------------------------------------------------------------------
# USAGE:
#      moduledoc.pl [-D] infile outfile
#      moduledoc.pl [-D] infile
#      moduledoc.pl [-d] outfile
#      moduledoc.pl [-h]
#
# Transfer electrical module definition documentation from one file to
# another, or just print the documentation in human-readable format.
# This program was written to maintain documentation in IBIS-macro files.
# See notes below about special comments.
#
#
# OPTIONS:
#
# -d                    Delete comments from the outfile. This is most
#                       useful for stripping comment blocks from a VA
#                       or VHD file so that it can be compared against
#                       a newer or older version of the file.
#
# -D                    Dump internal data structures to debug files:
#                       moduledoc_DEBUG_1.txt    After reading the input file
#                       moduledoc_DEBUG_2.txt    After reading the output file
#                       moduledoc_DEBUG_3.txt    After writing the output file
#
# -h                    Help (you are here).
#
#
# FILE ARGUMENTS:
#
# If only the infile argument is given, prints plain text documentation for
# all known modules in the infile to STDOUT.
#
# If the outfile argument is also given, module documentation info from the
# infile is annotated into the outfile. Only the documentation is transferred.
# The existing contents of the outfile are otherwise preserved.
#
# The infile and outfile arguments can be of the following types:
#
#    VA		Verilog-A file (can be infile or outfile)
#               As an infile, module names, terminal lists and parameter lists
#               are extracted from the code. Descriptive info is extracted from
#               existing comment blocks, if any.
#    VHD	VHDL-AMS file (can be infile or outfile)
#               TBD: not supported yet
#    XML	XML file  (can be infile or outfile)
#               Stores module docmentation info using a proprietary definition.
#               TBD: No DTD yet
#
# The file extensions must be one of .va, .vhd, or .xml to be recognized.
# Specified files must exist, with the exception that an XML outfile will
# be created if necessary. Since annotation is cumulative, an XML file will
# contain it's previous contents plus new contents when used as an existing
# outfile (unless -d is used, in which case the XML file will be emptied).
#
#
# SPECIAL COMMENTS
#
# Add a special comment containing '===modulelist===' to VA and VHD files
# to have a module listing annotated at that point. A VA example:
#
#//======================================================modulelist==========
#//==========================================================================
#
# Existing content between the two comment lines is replaced with the list.
# NOTE: the list will contain all known modules, even ones that are not
#       defined in the outfile.
#
# Add special 'PRIVATE' comments to parameters that should not be docuemented.
# A VA example:
#
#   parameter real M = Scale * Kval * sqrt(Lval_1 * Lval_2); // PRIVATE
#
# In the above case the parameter M will not appear in generated documentation.
#
#
# EXAMPLES:
#
#          To produce an XML documentation file from macrolib.va:
#
#              perl moduledoc.pl macrolib.va macrolib_doc.xml
#
#          To annotate documentation from an XML file into a new VA file:
#
#              perl moduledoc.pl macrolib_doc.xml new_macrolib.va
#
#          To annotate documentation from an XML file to a VHD file:
#
#              perl moduledoc.pl macrolib_doc.xml new_macrolib.vhd
#
#----------------------------------------------------------------------------

# DOCUMENTATION FORMATS:
#
# Module comment blocks in VA files look like this:
#
#// =========================================moduledoc======
#//
#// MODULE:      IBIS_R (p, n)
#//
#// DESCRIPTION:
#//              A simple resistor. The resistance value is calculated:
#//
#//              R = Rval * Scale
#//
#// TERMINALS:
#//
#//  p           positive resistor terminal
#//  n           negative resistor terminal
#//
#// PARAMETERS:
#//
#//  Rval        Resistance value
#//  Scale       Scaling factor for resistance value
#//
#// INSTANCE CALL EXAMPLES:
#//
#//  SPICE       X1 clk_n clkp IBIS_R Rval=100
#//
#//  VAMS        IBIS_R #(.Rval(100)) R1 (clk_n, clk_p);
#//
#// =========================================================
#
# The ===moduledoc=== comment is used by the annotator to recognize existing
# comment blocks in outfiles that are to be replaced.
#
# The corresponding info in XML looks like this:
#
#<MACRODOC>
#  <MODULE>
#    <MNAME>IBIS_R</MNAME>
#    <MDESC>
#      <P>A simple resistor. The resistance value is calculated:</P>
#      <P>R = Rval * Scale</P>
#    </MDESC>
#    <TERM>
#      <TNAME>p</TNAME>
#      <TDESC>positive resistor terminal</TDESC>
#    </TERM>
#    <TERM>
#      <TNAME>n</TNAME>
#      <TDESC>negative resistor terminal</TDESC>
#    </TERM>
#    <PARAM>
#      <PNAME>Rval</PNAME>
#      <PDESC>Resistance value</PDESC>
#      <PDEF>1.0</PDEF>
#    </PARAM>
#    <PARAM>
#      <PNAME>Scale</PNAME>
#      <PDESC>Scaling factor for resistance value</PDESC>
#      <PDEF>1.0</PDEF>
#    </PARAM>
#    <EXAMPLE>
#      <LANG>SPICE</LANG>
#      <CODE>X1 clk_n clkp IBIS_R Rval=100</CODE>
#    </EXAMPLE>
#    <EXAMPLE>
#      <LANG>VAMS</LANG>
#      <CODE>IBIS_R #(.Rval(100)) R1 (clk_n, clk_p);</CODE>
#    </EXAMPLE>
#  </MODULE>
#  <MODULE>
#    ...
#  </MODULE>
#</MACRODOC>

# Revision history:
#
# $Log: moduledoc.pl,v $
# Revision 1.10  2005/11/21 16:09:03  milabont
# Change VHDL-AMS extensin from VHDLA to VHD[D[D[D[D[D[D[D[D[D[D[D[D[D[D[D[D[D[D[Do
# Add support for annotating VHD files
#
# Revision 1.9  2005/11/19 22:12:47  milabont
# Fix backup file problems.
# Detect integer parameters.
#
# Revision 1.8  2005/11/11 22:12:00  milabont
# Retain the sequence of modules found in VA files, so that the module
# listing matches the module definition order.
# Backup XML files before writing.
#
# Revision 1.7  2005/11/07 18:01:15  milabont
# Add -d option to delete special comments.
# Add comments for more subroutines.
#
# Revision 1.6  2005/11/07 14:33:30  milabont
# Do not copy parameters marked PRIVATE in readVAcode().
#
# Revision 1.5  2005/11/04 17:44:06  milabont
# Stop using STDOUT for file output and STDERR for ordinary messages.
# Fix diagnostics.
# Print parameter default values.
#
# Revision 1.4  2005/11/04 15:06:09  milabont
# Make readVA() non-overwriting when storing info in $modules.
# It reports changes to STDERR instead. This change is not a complete
# solution, and more needs to be done. The problem was that info from
# the destination file was taking precedence over the source file.
#
# Revision 1.3  2005/11/03 20:41:16  milabont
# Fix handling of instance call examples.
# Ensure that comment blocks read from VA produce internal data
# that is identical to the internal data that printed them.
# Before this, annotating XML into a VA twice would produce
# two different files.
#
# Revision 1.2  2005/11/03 18:06:41  milabont
# Completely new argument format.
#
# Revision 1.1  2005/11/03 00:19:52  milabont
# Initial revision
#

use Getopt::Std qw(getopts);
use XML::Parser;
use Text::Wrap qw(wrap $columns $huge);
use Data::Dumper;

$Text::Wrap::unexpand = 0;
$columns = 75;
$huge = 'wrap';

getopts('dDh') || Usage();
Usage() if($opt_h);

# Map of legal file types, and whether they can be used for in and/or out.
my %ftypes = (
  VA	=> {
    	in	=> 1,
    	out	=> 1,
  },
  VHD	=> {
    	in	=> 0,
    	out	=> 1,
  },
  XML	=> {
    	in	=> 1,
    	out	=> 1,
  },
);

unless($opt_d) {
  $infile = shift;
  $inext = $infile;
  $inext =~ s/.*\.//;
  $intype = uc($inext);

  Usage() unless($infile ne '');

  die "Infile type $intype not supported" unless($ftypes{$intype}{in});

  die "Infile does not exist" unless(-r $infile);
}

$outfile = shift;
$outext = $outfile;
$outext =~ s/.*\.//;
$outtype = uc($outext);

if($outfile ne '') {
  die "Outfile type $outtype not supported" unless($ftypes{$outtype}{out});

  die "Outfile does not exist" unless(-r $outfile || $outtype eq 'XML');

  die "Infile and outfile must not be the same file" if($infile eq $outfile);
}

$modules = {};

# Extract documentation data from the infile.

unless($opt_d) {
  if($intype eq 'XML') {
    print "Reading XML file $infile\n";
    readXML($modules, $infile);
  }
  elsif($intype eq 'VA') {
    print "Reading code from VA file $file\n";
    readVAcode($modules, $infile);
    print "Reading comments from VA file $file\n";
    readVAcomments($modules, $infile);
  }
}

# Use "-D debugfile" to enable debug printing
printDebug('moduledoc_DEBUG_1.txt') if($opt_D);

if($outfile ne '') {

  unless($opt_d) {
    # Extract documentation data from the outfile.
    if($outtype eq 'XML') {
      if(-r $outfile) {
	print "Reading XML file $outfile\n";
	readXML($modules, $outfile)
      }
    }
    elsif($outtype eq 'VA') {
      print "Reading code from VA file $file\n";
      readVAcode($modules, $outfile);
      print "Reading comments from VA file $file\n";
      readVAcomments($modules, $outfile);
    }
  }

  # Use "-D debugfile" to enable debug printing
  printDebug('moduledoc_DEBUG_2.txt') if($opt_D);

  # See if there is any documentaion to work with.
  if(scalar(keys(%{$modules})) > 0 || $opt_d) {
    if($outtype eq 'XML') {
      my $srcfile = revRenameFile($outfile);
      print "File $outfile moved to $srcfile for backup\n" if($srcfile);
      print "Writing XML file $outfile\n";
      writeXML($modules, $outfile);
    }
    elsif($outtype eq 'VA') {
      my $srcfile = revRenameFile($outfile);
      print "File $outfile moved to $srcfile for backup\n" if($srcfile);
      print "Annotating $outtype file $outfile\n";
      my ($modcount, $listcount) = annotateVA($modules, $srcfile, $outfile);
      print "$modcount modules annotated\n";
      print "NOTE: no ===modulelist=== annotated\n" unless($listcount > 0);
    }
    elsif($outtype eq 'VHD') {
      my $srcfile = revRenameFile($outfile);
      print "File $outfile moved to $srcfile for backup\n" if($srcfile);
      print "Annotating $outtype file $outfile\n";
      my ($modcount, $listcount) = annotateVHD($modules, $srcfile, $outfile);
      print "$modcount modules annotated\n";
      print "NOTE: no ===modulelist=== annotated\n" unless($listcount > 0);
    }
  }
  else {
    die "No documentation data found in $infile or $outfile";
  }

  # Use "-D debugfile" to enable debug printing
  printDebug('moduledoc_DEBUG_3.txt') if($opt_D);
}
else {
  # No outfile; print each module to STDOUT
  open(DEST, '>-');
  foreach my $module (sort(keys(%{$modules}))) {
    printModule($module, '');
  }
  close(DEST);
}

exit 0;

#-------------------------------------------------------------------------
# readVAcode($modules, $file)
#
# Read a VA file, extracting module info from the actual code.
# Info is place in the $modules hash, nothing is returned.
#
# Existing info in the $modules hash takes precedence. This is done
# because the infile is read first, and that is the file from which
# new values are to be taken. This gives us a chances to note the changes.
# returns nothing.
#-------------------------------------------------------------------------

sub readVAcode {
  my $modules = shift;
  my $file = shift;
  my $module = '';
  my $pmode = 0;
  my $mcount = 0;
  my $tcount = 0;

  open(IN, "<$file");
  while(<IN>) {
    s/[\n\r]//g;
    # Strip off and save inline comments
    /\/\/(.*)/;
    my $comments = $1;
    s/\/\/.*//;
    # Note whether this line has a semi-colon, meaning it's the last param.
    my $semicol = /;/;
    # Clean up the line for easier parsing
    s/[,;]/ /g;
    s/=/ = /g;
    my @words = split();
    if($pmode) {
      # In parameter mode; every 'name = value' is a param.
      # Ignore parameters marked 'private'.
      if($words[1] eq '=' && $comments !~ /private/i) {
	my $param = $words[0];
	$param =~ s/[()]//g;
	my $pdef = join(' ', @words[2..$#words]);
	if($param ne '') {
	  if(exists($modules->{$module}{params}{$param}{pdef}) &&
	    $modules->{$module}{params}{$param}{pdef} ne '' &&
	    $modules->{$module}{params}{$param}{pdef} ne $pdef) {
	    printf ("Module %s : parameter %s default value changes from '%s' to '%s'\n",
	      $module, $param,
	      $pdef, $modules->{$module}{params}{$param}{pdef});
	  }
	  else {
	    $modules->{$module}{params}{$param}{pdef} = $pdef;
	  }
	  my $pseq = paramSeqNum($modules, $module, $param);
	  $modules->{$module}{params}{$param}{seq} = $pseq;
	}
      }
      $pmode = 0 if($semicol);
    }
    else {
      if($words[0] =~ /^module$/i) {
	shift @words;
	$module = uc($words[0]);
	$modules->{$module}{seq} = ++$mcount;
	$tcount = 0;
	foreach my $term (@words[1..$#words]) {
	  ++$tcount;
	  $term =~ s/[()]//g;
	  my $tseq = termSeqNum($modules, $module, $term);
	  if(exists($modules->{$module}{terms}{$term})) {
	    unless($tcount == $tseq) {
	      printf ("Module %s : terminal %s moves from position %d to %d\n",
	        $module, $term, $tcount, $tseq);
	    }
	  }
	  $modules->{$module}{terms}{$term}{seq} = $tseq;
	}
      }
      elsif($words[0] =~ /^parameter$/i && 
            ($words[1] =~ /^real$/i || $words[1] =~ /^integer$/i) &&
            $comments !~ /private/i) {
	my $param = $words[2];
	$param =~ s/[()]//g;
	my $pdef = join(' ', @words[4..$#words]);
	if($param ne '') {
	  if(exists($modules->{$module}{params}{$param}{pdef}) &&
	    $modules->{$module}{params}{$param}{pdef} ne '' &&
	    $modules->{$module}{params}{$param}{pdef} ne $pdef) {
	    printf ("Module %s : parameter %s default value changes from '%s' to '%s'\n",
	      $module, $param,
	      $pdef, $modules->{$module}{params}{$param}{pdef});
	  }
	  else {
	    $modules->{$module}{params}{$param}{pdef} = $pdef;
	  }
	  my $pseq = paramSeqNum($modules, $module, $param);
	  $modules->{$module}{params}{$param}{seq} = $pseq;
	}
	else {
	  $pmode = 1;
	}
      }
    }
  }
  close(IN);
}

#-------------------------------------------------------------------------
# readVAcomments($modules, $file)
#
# Read a VA file, extracting module info from existing comment blocks.
# Info is place in the $modules hash, nothing is returned.
#
# Existing info in the $modules hash takes precedence. This is done
# because the infile is read first, and that is the file from which
# new values are to be taken. This gives us a chances to note the changes.
# Returns nothing.
#-------------------------------------------------------------------------

sub readVAcomments {
  my $modules = shift;
  my $file = shift;
  my @keys = (
    'MODULE:',
    'DESCRIPTION:',
    'TERMINALS:',
    'PARAMETERS:',
    'INSTANCE CALL EXAMPLES:',
    );
  my $module = '';
  my $curmdesc = '';
  my $curlang = '';

  open(IN, "<$file");
  while(<IN>) {
    chomp;
    next unless(/^\/\//);
    s/^\/+\s*//;
    my @words = split();
    # See if this line looks like one of the documentation keyword lines.
    my $keyword = '';
    foreach my $key (@keys) {
      $keyword = $key if($_ =~ /$key/);
    }

    if(/=================/) {
      $section = '';
    }

    # The MODULE: line has the module name on it.
    # The DESCRRIPTION: needs to be flushed once we reached another keyword.
    if($keyword) {
      if($keyword eq 'MODULE:') {
	$module = $words[1];
      }
      if($keyword ne 'DESCRIPTION:' && $curmdesc =~ /[^\s]/) {
	my $pcount = 0;
	foreach my $para (split(/\n\n/, $curmdesc)) {
	  $modules->{$module}{mdesc}{++$pcount} = $para if($para =~ /[^\s]/);
	}
	$curmdesc = '';
      }
      $section = $keyword;
    }
    else {
      if($section eq 'TERMINALS:' && $#words > 0) {
	$words[$#words] =~ s/\.$//;
	$term = $words[0];
	$tdesc = join(' ', @words[1..$#words]);
	# If seq has already been set by readVAcode(), this will make no change.
	$modules->{$module}{terms}{$term}{seq} =
	  termSeqNum($modules, $module, $term);
	$modules->{$module}{terms}{$term}{tdesc} = $tdesc;
      }
      elsif($section eq 'PARAMETERS:' && $#words > 0) {
	$words[$#words] =~ s/\.$//;
	if($words[0] eq 'DEFAULT' && $words[2] eq '=') {
	  $param = $words[1];
	  $pdef = join(' ', @words[3..$#words]);
	  $pdef =~ s/'//g;
	  $modules->{$module}{params}{$param}{pdef} = $pdef if($param);
	}
	else {
	  $param = $words[0];
	  $pdesc = join(' ', @words[1..$#words]);
	  # If seq has already been set by readVAcode(), this will make no change.
	  $modules->{$module}{params}{$param}{seq} =
	    paramSeqNum($modules, $module, $param);
	  $modules->{$module}{params}{$param}{pdesc} = $pdesc;
	}
      }
      elsif($section eq 'INSTANCE CALL EXAMPLES:' && $#words >= 0) {
	$lang = $words[0];
	if($lang =~ /[A-Z]+:$/) {
	  $curlang = $lang;
	  $curlang =~ s/://;
	  $code = $_;
	  $code =~ s/^\s*$lang\s*//;
	  $modules->{$module}{examples}{$curlang}{code} = $code;
	}
	else {
	  $code = $_;
	  $code =~ s/^\s*//;
	  $modules->{$module}{examples}{$curlang}{code} .= "\n$code";
	}
      }
      elsif($section eq 'DESCRIPTION:') {
	my $desc = $_;
	$desc =~ s/^\s*//;
	$curmdesc .= "$desc\n";
      }
    }
  }
  close(IN);

  if($curmdesc =~ /[^\s]/) {
    my $pcount = 0;
    foreach my $para (split(/\n\n/, $curmdesc)) {
      $modules->{$module}{mdesc}{++$pcount} = $para;
    }
  }
}

#-------------------------------------------------------------------------
#
# paramSeqNum($modules, $module, $param)
#
# Determine the positional sequence number for a named parameter.
# If the parameter already has a sequence number in the $modules hash,
# that is returned. Otherwise the lowest number not taken by any existing
# parameter is returned. Sequence numbers start at 1.
# Returns the sequence number.
#
#-------------------------------------------------------------------------

sub paramSeqNum {
  my $modules = shift;
  my $module = shift;
  my $param = shift;
  my $seq = 1;

  if(exists($modules->{$module}{params}{$param}{seq})) {
    $seq = $modules->{$module}{params}{$param}{seq};
  }
  else {
    foreach my $p (keys(%{$modules->{$module}{params}})) {
      if(exists($modules->{$module}{params}{$p}{seq}) &&
         $modules->{$module}{params}{$p}{seq} >= $seq) {
	$seq = $modules->{$module}{params}{$p}{seq} + 1;
      }
    }
  }
  return $seq;
}

#-------------------------------------------------------------------------
#
# termSeqNum($modules, $module, $term)
#
# Determine the positional sequence number for a named terminal.
# If the terminal already has a sequence number in the $modules hash,
# that is returned. Otherwise the lowest number not taken by any existing
# terminal is returned. Sequence numbers start at 1.
# Returns the sequence number.
#
#-------------------------------------------------------------------------

sub termSeqNum {
  my $modules = shift;
  my $module = shift;
  my $term = shift;
  my $seq = 1;

  if(exists($modules->{$module}{terms}{$term}{seq})) {
    $seq = $modules->{$module}{terms}{$term}{seq};
  }
  else {
    foreach my $t (keys(%{$modules->{$module}{terms}})) {
      if(exists($modules->{$module}{terms}{$t}{seq}) &&
         $modules->{$module}{terms}{$t}{seq} >= $seq) {
	$seq = $modules->{$module}{terms}{$t}{seq} + 1;
      }
    }
  }
  return $seq;
}

#-------------------------------------------------------------------------
#
# annotateVA($modules, $srcfile, $outfile)
#
# Copy a source Verilog-A file to an output file, adding module comments.
# Existing special comment blocks are removed first.
# Returns a list of 2 numbers:
#  number of module definitions commented.
#  number of modules listed
#
#-------------------------------------------------------------------------

sub annotateVA {
  my $modules = shift;
  my $srcfile = shift;
  my $outfile = shift;
  my $prefix = '//';
  my $modcount = 0;
  my $listcount = 0;

  open(SRC, "<$srcfile") || die "Unable to open $srcfile for read";
  my @lines = (<SRC>);
  close(SRC);

  my @lines2 = ();
  my $indocblock = 0;

  # Search for existing comment blocks. Don't copy them.
  foreach my $line (@lines) {
    if($line =~ /^$prefix.*===moduledoc===/) {
      $indocblock = 1;
    }
    elsif($indocblock && $line =~ /^$prefix.*======/) {
      $indocblock = 0;
      next;
    }
    push(@lines2, $line) unless($indocblock);
    if($line =~ /^$prefix.*===modulelist===/) {
      $indocblock = 1;
    }
  }

  # Print the remaining lines, inserting comment blocks as needed.
  open(DEST, ">$outfile") || die "Unable to open $outfile for write";
  foreach my $line (@lines2) {
    if($line =~ /^module\s+([^\s(]*)/i) {
      my $module = $1;
      $modcount += printModule($module, $prefix);
    }
    elsif($line =~ /^$prefix.*===modulelist===/) {
      $listcount += printModuleList($prefix);
      next;
    }
    print DEST $line;
  }
  close(DEST);
  return ($modcount, $listcount);
}

#-------------------------------------------------------------------------
#
# annotateVHD($modules, $srcfile, $outfile)
#
# Copy a source Verilog-A file to an output file, adding module comments.
# Existing special comment blocks are removed first.
# Returns a list of 2 numbers:
#  number of module definitions commented.
#  number of modules listed
#
#-------------------------------------------------------------------------

sub annotateVHD {
  my $modules = shift;
  my $srcfile = shift;
  my $outfile = shift;
  my $prefix = '--';
  my $modcount = 0;
  my $listcount = 0;

  open(SRC, "<$srcfile") || die "Unable to open $srcfile for read";
  my @lines = (<SRC>);
  close(SRC);

  my @lines2 = ();
  my $indocblock = 0;

  # Search for existing comment blocks. Don't copy them.
  foreach my $line (@lines) {
    if($line =~ /^$prefix.*===moduledoc===/) {
      $indocblock = 1;
    }
    elsif($indocblock && $line =~ /^$prefix.*======/) {
      $indocblock = 0;
      next;
    }
    push(@lines2, $line) unless($indocblock);
    if($line =~ /^$prefix.*===modulelist===/) {
      $indocblock = 1;
    }
  }

  # Print the remaining lines, inserting comment blocks as needed.
  open(DEST, ">$outfile") || die "Unable to open $outfile for write";
  foreach my $line (@lines2) {
    if($line =~ /^\s*entity\s+([^\s(]*)/i) {
      my $module = $1;
      $modcount += printModule($module, $prefix);
    }
    elsif($line =~ /^$prefix.*===modulelist===/) {
      $listcount += printModuleList($prefix);
      next;
    }
    print DEST $line;
  }
  close(DEST);
  return ($modcount, $listcount);
}


#-------------------------------------------------------------------------
#
# printModule($module, $prefix)
#
# Print a comment block for one module.
# Each line begins with the given prefix.
# The DEST file descriptor must be opened for output before calling.
# Returns 1 if the specified module is printed.
#
#-------------------------------------------------------------------------

sub printModule {
  my $module = shift;
  my $prefix = shift;
  my $count = 0;
  if(exists($modules->{$module})) {
    ++$count;
    my $mdesc = '';
    if(exists($modules->{$module}{mdesc})) {
      my $leader = "$prefix              ";
      my @descparas = map({$modules->{$module}{mdesc}{$_};}
	  sort(keys(%{$modules->{$module}{mdesc}})));
      $mdesc = join("\n$prefix\n",
        map({s/[\n\r]/ /g; s/  / /g; wrap($leader, $leader, $_);} @descparas));
    }
    # Warn if the description is empty.
    unless($mdesc =~ /[^\s]/) {
      print STDERR "WARNING: No module description for $module\n";
    }
    # Make a text list of terminals.
    my $tlist = join(', ', sort(
	{$modules->{$module}{terms}{$a}{seq} <=> $modules->{$module}{terms}{$b}{seq};} 
	keys(%{$modules->{$module}{terms}})));

    print DEST << "EOF";
$prefix =========================================moduledoc======
$prefix
$prefix MODULE:      $module ($tlist)
$prefix
$prefix DESCRIPTION:
$mdesc
$prefix
$prefix TERMINALS:
$prefix
EOF
    foreach my $term (sort(
	{$modules->{$module}{terms}{$a}{seq} <=> $modules->{$module}{terms}{$b}{seq};} 
	keys(%{$modules->{$module}{terms}}))) {
      my $tdesc = $modules->{$module}{terms}{$term}{tdesc};
      printf DEST ("%s  %-10s  %s\n", $prefix, $term, $tdesc);
    }
    print DEST << "EOF";
$prefix
$prefix PARAMETERS:
$prefix
EOF
    foreach my $param (sort(
	{$modules->{$module}{params}{$a}{seq} <=> $modules->{$module}{params}{$b}{seq};} 
        keys(%{$modules->{$module}{params}}))) {
      my $pdesc = $modules->{$module}{params}{$param}{pdesc};
      my $pdef = $modules->{$module}{params}{$param}{pdef};
      printf DEST ("%s  %-10s  %s\n", $prefix, $param, $pdesc);
      printf DEST ("%s  %-10s  DEFAULT %s = '%s'\n", $prefix, '', $param, $pdef);
      printf DEST "$prefix\n";
    }
    print DEST << "EOF";
$prefix
$prefix INSTANCE CALL EXAMPLES:
EOF
    foreach my $lang (sort(keys(%{$modules->{$module}{examples}}))) {
      my $code = $modules->{$module}{examples}{$lang}{code};
      my $l1 = sprintf("%s  %-11s ", $prefix, $lang.':');
      my $l2 = "$prefix              ";
      my $example = wrap($l1, $l2, $code);
      print DEST "$prefix\n$example\n";
    }
    print DEST << "EOF";
$prefix
$prefix =========================================================
EOF
  }
  return $count;
}


#-------------------------------------------------------------------------
#
# printModuleList($prefix)
#
# Print a comment block listing all module calls.
# Each line begins with the given prefix.
# The DEST file descriptor must be opened for output before calling.
# All known modules are listed, even ones not found in the outfile.
# Returns the number of modules listed.
#
#-------------------------------------------------------------------------

sub printModuleList {
  my $prefix = shift;
  print DEST << "EOF";
$prefix =========================================modulelist=====
$prefix
EOF
  my $count = scalar(keys(%{$modules}));
  if($count > 0) {
    print DEST << "EOF";
$prefix LIST OF MODULES:
$prefix
EOF
  }
  foreach my $module (sort(
    {$modules->{$a}{seq} <=> $modules->{$b}{seq};}
      keys(%{$modules}))) {
    my $tlist = join(', ',
      sort(
	{$modules->{$module}{terms}{$a}{seq} <=> $modules->{$module}{terms}{$b}{seq};} 
	keys(%{$modules->{$module}{terms}})));
    printf DEST ("%s   %-20s (%s)\n", $prefix, $module, $tlist)
  }
  print DEST << "EOF";
$prefix
$prefix =========================================================
EOF
  return $count;
}


#-------------------------------------------------------------------------
#
# printDebug($dbgfile)
#
# Print the contents of the $modules hash for debug.
# Returns nothing.
#
#-------------------------------------------------------------------------

sub printDebug {
  my $dbgfile = shift;
  open(DBG, ">$dbgfile");
  my %modulesh = %{$modules};
  foreach my $infodumper (
      Data::Dumper->new([\%modulesh], ["modules"]),
    ) {
    $infodumper->Indent(1);
    $infodumper->Purity(1);
    print DBG $infodumper->Dump();
  }
  close(DBG);
}

#-------------------------------------------------------------------------
#
# readXML($modules, $xmlfile)
#
# Read the data inan XML file into the $modules hash.
# Returns nothing.
#
#-------------------------------------------------------------------------

sub readXML {
  my $modules = shift;
  my $xmlfile = shift;

  $curmodule = '';
  $curmdesc = '';
  $curterm = '';
  $curtdesc = '';
  $curparam = '';
  $curpdesc = '';
  $curpdef = '';
  $curlang = '';
  $curcode = '';
  $pcount = 0;
  $paramcount = 0;
  $termcount = 0;
  $examplecount = 0;

  $xml = new XML::Parser(Style => 'Stream');
  $xml->parsefile($xmlfile) || die "Failed parsing $xmlfile";
}

########################################################################
#
#  Stream handler routines called by XML::Parser
#  These build the $modules hash.
#
########################################################################

# Module state variables

$curmodule = '';
$curmdesc = '';
$curterm = '';
$curtdesc = '';
$curparam = '';
$curpdesc = '';
$curpdef = '';
$curlang = '';
$curcode = '';
$pcount = 0;
$paramcount = 0;
$termcount = 0;
$examplecount = 0;

sub StartDocument {
  return 1;
}

sub StartTag {
  push(@tagstack, $_);
  if($_ eq '<MODULE>') {
    $curmdesc = '';
    $curterm = '';
    $curtdesc = '';
    $curparam = '';
    $curpdesc = '';
    $curpdef = '';
    $curlang = '';
    $curcode = '';
    $pcount = 0;
    $paramcount = 0;
    $termcount = 0;
    $examplecount = 0;
  }
  elsif($_ eq '<P>') {
    ++$pcount;
  }
  elsif($_ eq '<PARAM>') {
    ++$paramcount;
  }
  elsif($_ eq '<TERM>') {
    ++$termcount;
  }
  elsif($_ eq '<EXAMPLE>') {
    ++$examplecount;
  }
}

sub EndTag {
  my $tag = pop(@tagstack);
  if($tag eq '<P>') {
    if($curmodule ne '') {
      $modules->{$curmodule}{mdesc}{$pcount} = $curmdesc;
    }
    else {
      print STDERR "No module set for <MDESC><P>\n";
    }
  }
  elsif($tag eq '<TERM>') {
    if($curmodule ne '') {
      $modules->{$curmodule}{terms}{$curterm}{tdesc} = $curtdesc;
      $modules->{$curmodule}{terms}{$curterm}{seq} = $termcount;
    }
    else {
      print STDERR "No module set for <TERM>\n";
    }
  }
  elsif($tag eq '<PARAM>') {
    if($curmodule ne '') {
      $modules->{$curmodule}{params}{$curparam}{pdef} = $curpdef;
      $modules->{$curmodule}{params}{$curparam}{pdesc} = $curpdesc;
      $modules->{$curmodule}{params}{$curparam}{seq} = $paramcount;
    }
    else {
      print STDERR "No module set for <PARAM>\n";
    }
  }
  elsif($tag eq '<EXAMPLE>') {
    if($curmodule ne '') {
      $modules->{$curmodule}{examples}{$curlang}{code} = $curcode;
    }
    else {
      print STDERR "No module set for <EXAMPLE>\n";
    }
  }
}

sub Text {
  s/[\n\r]*$//g;
  if($tagstack[$#tagstack] eq '<MNAME>') {
    $curmodule = $_;
  }
  elsif($tagstack[$#tagstack] eq '<P>') {
    $curmdesc = $_;
  }
  elsif($tagstack[$#tagstack] eq '<PNAME>') {
    $curparam = $_;
  }
  elsif($tagstack[$#tagstack] eq '<PDESC>') {
    $curpdesc = $_;
  }
  elsif($tagstack[$#tagstack] eq '<PDEF>') {
    $curpdef = $_;
  }
  elsif($tagstack[$#tagstack] eq '<TNAME>') {
    $curterm = $_;
  }
  elsif($tagstack[$#tagstack] eq '<TDESC>') {
    $curtdesc = $_;
  }
  elsif($tagstack[$#tagstack] eq '<LANG>') {
    $curlang = $_;
  }
  elsif($tagstack[$#tagstack] eq '<CODE>') {
    $curcode = $_;
  }
}

sub PI {
  return 1;
}

sub EndDocument {
  return 1;
}

########################################################################
#  END OF Stream handler routines
########################################################################

#-------------------------------------------------------------------------
#
# writeXML($modules, $xmlfile)
#
# Write the data in the $modules hash to an XML file.
# Returns nothing.
#
#-------------------------------------------------------------------------

sub writeXML {
  my $modules = shift;
  my $xmlfile = shift;
  open(XML, ">$xmlfile");
  print XML << "EOF";
<MACRODOC>
EOF
  foreach my $module (sort(
      {$modules->{$a}{seq} <=> $modules->{$b}{seq};}
      keys(%{$modules}))) {
    print XML << "EOF";
  <MODULE>
    <MNAME>$module</MNAME>
    <MDESC>
EOF
    foreach my $para (sort(keys(%{$modules->{$module}{mdesc}}))) {
      print XML << "EOF";
      <P>$modules->{$module}{mdesc}{$para}</P>
EOF
    }
    print XML << "EOF";
    </MDESC>
EOF
    foreach my $term (sort(
	{$modules->{$module}{terms}{$a}{seq} <=> $modules->{$module}{terms}{$b}{seq};} 
	keys(%{$modules->{$module}{terms}}))) {
      print XML << "EOF";
    <TERM>
      <TNAME>$term</TNAME>
      <TDESC>$modules->{$module}{terms}{$term}{tdesc}</TDESC>
    </TERM>
EOF
    }
    foreach my $param (sort(
	{$modules->{$module}{params}{$a}{seq} <=> $modules->{$module}{params}{$b}{seq};} 
	keys(%{$modules->{$module}{params}}))) {
      print XML << "EOF";
    <PARAM>
      <PNAME>$param</PNAME>
      <PDESC>$modules->{$module}{params}{$param}{pdesc}</PDESC>
      <PDEF>$modules->{$module}{params}{$param}{pdef}</PDEF>
    </PARAM>
EOF
    }
    foreach my $lang (sort(keys(%{$modules->{$module}{examples}}))) {
      print XML << "EOF";
    <EXAMPLE>
      <LANG>$lang</LANG>
      <CODE>$modules->{$module}{examples}{$lang}{code}</CODE>
    </EXAMPLE>
EOF
    }
    print XML << "EOF";
  </MODULE>
EOF
  }
  print XML << "EOF";
</MACRODOC>
EOF
  close(XML);
}


#-------------------------------------------------------------------------
#
# revRenameFile($file)
#
# Rename a file to preserve it. Revision-numbered names like file,1.ibs
# file,2.ibs are chosen, up to a set limit (currently 4).
# Returns the name of the file containing the original data, which is
# most likely not the file name passed in..
#
#-------------------------------------------------------------------------

sub revRenameFile {
  my $file = shift;
  $file =~ /(.*)\.(.*)$/;
  my $root = $1;
  my $ext = $2;
  my $origfile = $file;
  my $renamefile = '';

  my $i = 4;
  while(--$i >= 0) {
    my $j = $i + 1;
    my $fromfile = $i > 0 ? "$root,$i.$ext" : $file;
    my $tofile = "$root,$j.$ext";
    if(-f $fromfile) {
      rename($fromfile, $tofile);
      $renamefile = $tofile if($fromfile eq $file);
    }
  }
  return $renamefile;
}

#-------------------------------------------------------------------------
#
# Usage()
#
# Description:  This function will display the usage message which is
#               embedded in the header comment for this file.
#               No return from this, because exit() is called.
#
#-------------------------------------------------------------------------

sub Usage {

  my $Display = 0;

  open(THISFILE, "< $0");

  while (<THISFILE>) {
    if (/-------------------/) {
      if ($Display) {
	exit 1;
      } else {
        $Display = 1;
      }
    } elsif ($Display) {
      s/^#//;
      print STDERR;
    }
  }
}
